from typing import Sequence
import sympy as sp

from quple import TemplateCircuitBlock


class PauliBlock(TemplateCircuitBlock):
    def __init__(self, pauli_string:str):
        self.paulis = pauli_string[::-1]
        self.indices = [i for i, pauli in enumerate(self.paulis) if pauli != 'I']
        self._num_block_qubits = len(self.paulis)
    
    @staticmethod
    def change_basis(circuit:'quple.EncodingCircuit', qubits:Sequence[int],
                     pauli_string:Sequence[str], inverse=False) -> None:
        for i, pauli in enumerate(pauli_string):
            if pauli == 'X':
                circuit.H(qubits[i])
            elif pauli == 'Y':
                circuit.RX(-np.pi / 2 if inverse else np.pi / 2, qubits[i])            

    def build(self, circuit:'quple.EncodingCircuit', qubits:Sequence[int]):
        if not self.indices:
            return None
        
        PauliBlock.change_basis(circuit, qubits, self.paulis)
        qubits_to_entangle = tuple(qubits[i] for i in self.indices)
        encoded_value = circuit.encode_parameters(list(qubits))
        circuit.entangle(qubits_to_entangle)    
        circuit.RZ(sp.pi*encoded_value, qubits[-1])
        circuit.entangle(qubits_to_entangle, inverse=True)

        PauliBlock.change_basis(circuit, qubits, self.paulis, inverse=True)       
    
    @property
    def num_block_qubits(self) -> int:
        return self._num_block_qubits


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="quple", # Replace with your own username
    version="0.6.0.1",
    author="Alkaid Cheng",
    author_email="chi.lung.cheng@cern.ch",
    description="A framework for quantum machine learning in high energy physics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.cern.ch/clcheng/quple",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[   
          'cirq',
          'numpy',
          'tensorflow==2.1.0',
          'tensorflow_quantum==0.3.0'
      ],
    python_requires='>=3.5',
)
